/*
	modeles glest g3d
	http://www.glest.org
	
	mailto:jciehl@bat710.univ-lyon1.fr
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <GL/glut.h>

#include "g3dmodel.h"
#include "vec.h"
#include "img.h"
#include "tga.h"



void g3dmesh_anim_init(G3DMESH *mesh)
{
	if(mesh->vertexFrameCount > 1)
	{
		mesh->anim_vertices= (VEC3 *) malloc(sizeof(VEC3) * mesh->vertexFrameCount * mesh->pointCount);
		assert(mesh->anim_vertices!=NULL);
	}
	else
		mesh->anim_vertices= mesh->vertices;
	
	if(mesh->normalFrameCount > 1)
	{
		mesh->anim_normals= (VEC3 *) malloc(sizeof(VEC3) * mesh->normalFrameCount * mesh->pointCount);
		assert(mesh->anim_normals!=NULL);
	}
	else
		mesh->anim_normals= mesh->normals;
	
	if(mesh->texCoordFrameCount > 1)
	{
		mesh->anim_texCoords= (VEC2 *) malloc(sizeof(VEC2) * mesh->texCoordFrameCount * mesh->pointCount);
		assert(mesh->anim_texCoords!=NULL);
	}
	else
		mesh->anim_texCoords= mesh->texCoords;

	if(mesh->colorFrameCount > 1)	
	{
		mesh->anim_colors= (VEC4 *) malloc(sizeof(VEC4) * mesh->colorFrameCount);
		assert(mesh->anim_colors!=NULL);
	}
	else
		mesh->anim_colors= mesh->colors;
		
}


void g3dtexture_load(char *path, char *texName, G3DMODEL *model, int imesh)
{
	char texfname[PATH_MAX];
	G3DMESH *mesh;
	IMG *tex;
	int i;

	mesh= &model->meshes[imesh];
	mesh->texName[0]= 0;
	mesh->tex= 0;
	
	if(texName[0]!=0)
	{
		printf("  texture %d '%s'...", imesh, texName);

		// recherche parmi les textures deja chargees
		for(i= 0; i < imesh; i++)
			if(strcmp(texName, model->meshes[i].texName)==0)
				break;
		
		if(i < imesh)
		{
			// texture deja presente
			mesh->tex= model->meshes[i].tex;
			printf(" done.  link to texture %d.\n", i);

		}
		else
		{
			// charger la texture
			sprintf(texfname, "%s%s", path, texName);
			tex= lire_tga(texfname);
			if(tex!=NULL)
			{
				strcpy(mesh->texName, texName);
				
				// ecrire_tga("snap.tga", tex);
				
				glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
				glGenTextures(1, &mesh->tex);
				assert(glGetError()==GL_NO_ERROR);
				
				glBindTexture(GL_TEXTURE_2D, mesh->tex);
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
				assert(glGetError()==GL_NO_ERROR);
			
				if(tex->bits==24)
				{
					printf(" bind %u", mesh->tex);
					
					glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB,
						tex->largeur, tex->hauteur, 0,
						GL_RGB, GL_UNSIGNED_BYTE, tex->data);
					
					assert(glGetError()==GL_NO_ERROR);
				}
				else if(tex->bits==32)
				{
					printf(" bind %u", mesh->tex);
					
					glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA,
						tex->largeur, tex->hauteur, 0,
						GL_RGBA, GL_UNSIGNED_BYTE, tex->data);
					
					assert(glGetError()==GL_NO_ERROR);
				}
				else
				{
					printf(" delete");
					glDeleteTextures(1, &mesh->tex);
					mesh->texName[0]= 0;
					mesh->tex= 0;
	
					assert(glGetError()==GL_NO_ERROR);
				}
				
				printf(" done. %dx%d %d bits.\n", tex->largeur, tex->hauteur, tex->bits);
				free_img(tex);
			}
			else
			{
				printf(" failed.\n");
				mesh->texName[0]= 0;
				mesh->tex= 0;
			}
		}
	}
}	


void g3dmesh_loadv2(G3DMESH *mesh, char *path, FILE *f, G3DMODEL *model, int imesh)
{
	// read header
	G3DMESH_HEADERV2 meshHeader;
	
	fread(&meshHeader, sizeof(meshHeader), 1, f);
	
	// init
	mesh->vertexFrameCount= meshHeader.vertexFrameCount;
	mesh->normalFrameCount= meshHeader.normalFrameCount;
	mesh->texCoordFrameCount= meshHeader.texCoordFrameCount;
	mesh->colorFrameCount= meshHeader.colorFrameCount;
	mesh->pointCount= meshHeader.pointCount;
	mesh->indexCount= meshHeader.indexCount;

	printf("  mesh %d v2\n", imesh);
	printf("      frameCount vertex %d, normal %d, texCoord %d, color %d\n", 
		mesh->vertexFrameCount,
		mesh->normalFrameCount,
		mesh->texCoordFrameCount,
		mesh->colorFrameCount);
	printf("      pointCount %d, indexCount %d\n",
		mesh->pointCount,
		mesh->indexCount);

	// allocation
	mesh->vertices= (VEC3 *) malloc(sizeof(VEC3) * meshHeader.vertexFrameCount * meshHeader.pointCount);
	assert(mesh->vertices!=NULL);
	
	mesh->normals= (VEC3 *) malloc(sizeof(VEC3) * meshHeader.normalFrameCount * meshHeader.pointCount);
	assert(mesh->normals!=NULL);
	
	mesh->texCoords= (VEC2 *) malloc(sizeof(VEC2) * meshHeader.texCoordFrameCount * meshHeader.pointCount);
	assert(mesh->texCoords!=NULL);
	
	mesh->colors= (VEC4 *) malloc(sizeof(VEC4) * meshHeader.colorFrameCount);
	assert(mesh->colors!=NULL);
	
	mesh->indices= (int *) malloc(sizeof(int) * meshHeader.indexCount);
	assert(mesh->indices!=NULL);
	
	// 
	mesh->twoSided= 0;
	mesh->customTexture= -1;
	
	// charger la texture du maillage
	g3dtexture_load(path, meshHeader.texName, model, imesh);

	// charger toutes les donnees
	// fix verifier que tout est bien lu !!
	fread(mesh->vertices, sizeof(VEC3) * meshHeader.vertexFrameCount * meshHeader.pointCount, 1, f);
	fread(mesh->normals, sizeof(VEC3) * meshHeader.normalFrameCount * meshHeader.pointCount, 1, f);
	fread(mesh->texCoords, sizeof(VEC2) * meshHeader.texCoordFrameCount * meshHeader.pointCount, 1, f);
	fread(mesh->colors, sizeof(VEC4) * meshHeader.colorFrameCount, 1, f);
	fread(mesh->indices, sizeof(int) * meshHeader.indexCount, 1, f);
}

void g3dmesh_load(G3DMESH *mesh, char *path, FILE *f, G3DMODEL *model, int imesh)
{
	G3DMESH_HEADER meshHeader;
	
	// charger l'entete
	fread(&meshHeader, sizeof(meshHeader), 1, f);
	
	// init
	mesh->vertexFrameCount= meshHeader.vertexFrameCount;
	mesh->normalFrameCount= meshHeader.normalFrameCount;
	mesh->texCoordFrameCount= meshHeader.texCoordFrameCount;
	mesh->colorFrameCount= meshHeader.colorFrameCount;
	mesh->pointCount= meshHeader.pointCount;
	mesh->indexCount= meshHeader.indexCount;

	printf("  mesh %d v1\n", imesh);
	printf("      frameCount vertex %d, normal %d, texCoord %d, color %d\n", 
		mesh->vertexFrameCount,
		mesh->normalFrameCount,
		mesh->texCoordFrameCount,
		mesh->colorFrameCount);
	printf("      pointCount %d, indexCount %d\n",
		mesh->pointCount,
		mesh->indexCount);

	// allocation
	mesh->vertices= (VEC3 *) malloc(sizeof(VEC3) * meshHeader.vertexFrameCount * meshHeader.pointCount);
	assert(mesh->vertices!=NULL);
	
	mesh->normals= (VEC3 *) malloc(sizeof(VEC3) * meshHeader.normalFrameCount * meshHeader.pointCount);
	assert(mesh->normals!=NULL);
	
	mesh->texCoords= (VEC2 *) malloc(sizeof(VEC2) * meshHeader.texCoordFrameCount * meshHeader.pointCount);
	assert(mesh->texCoords!=NULL);
	
	mesh->colors= (VEC4 *) malloc(sizeof(VEC4) * meshHeader.colorFrameCount);
	assert(mesh->colors!=NULL);
	
	mesh->indices= (int *) malloc(sizeof(int) * meshHeader.indexCount);
	assert(mesh->indices!=NULL);
	
	//
	mesh->twoSided= 0;
	mesh->customTexture= -1;
	
	// read texture
	g3dtexture_load(path, meshHeader.texName, model, imesh);
	
	// read data
	fread(mesh->vertices, sizeof(VEC3) * meshHeader.vertexFrameCount * meshHeader.pointCount, 1, f);
	fread(mesh->normals, sizeof(VEC3) * meshHeader.normalFrameCount * meshHeader.pointCount, 1, f);
	fread(mesh->texCoords, sizeof(VEC2) * meshHeader.texCoordFrameCount * meshHeader.pointCount, 1, f);
	fread(mesh->colors, sizeof(VEC4) * meshHeader.colorFrameCount, 1, f);
	fread(mesh->indices, sizeof(int) * meshHeader.indexCount, 1, f);
}



G3DMODEL *g3dmodel_load(char *pathname)
{
	char path[PATH_MAX];
	G3DMODEL_HEADER modelHeader;
	G3DMODEL *model;
	FILE *f;
	char *base;
	int i;

	strcpy(path, pathname);
	base= strrchr(path, '/');
	if(base!=NULL)
		base[1]= 0;
	else
		path[0]= 0;
	
	f= fopen(pathname,"rb");
	if(f==NULL) 
	{
		fprintf(stderr, "error opening g3d model file '%s'\n", path);
		return NULL;
	}

	fread(&modelHeader, sizeof(modelHeader), 1, f);

	//	
	model= g3dmodel_new();
	model->meshCount= modelHeader.meshCount;

	if(modelHeader.id[0]!='G' || modelHeader.id[1]!='3' || modelHeader.id[2]!='D')
	{
		fprintf(stderr, "'%s' is not a valid g3d model\n", path);
		g3dmodel_free(model);
		return NULL;
	}

	switch(modelHeader.version)
	{
		case 2:
			// version 2 du maillage
			model->meshes= (G3DMESH *) malloc(sizeof(G3DMESH) * modelHeader.meshCount);
			for(i= 0; i < model->meshCount; i++)
			{
				g3dmesh_loadv2(&model->meshes[i], path, f, model, i);
				g3dmesh_anim_init(&model->meshes[i]);
			}
		break;
	
		case 3:
			// version 3 du maillage
			model->meshes= (G3DMESH *) malloc(sizeof(G3DMESH) * modelHeader.meshCount);
			for(i= 0; i < model->meshCount; ++i)
			{
				g3dmesh_load(&model->meshes[i], path, f, model, i);
				g3dmesh_anim_init(&model->meshes[i]);
			}
		break;
	
		default:
			fprintf(stderr, "'%s' is not a valid g3d model\n", path);
			fclose(f);
			g3dmodel_free(model);
			return NULL;
	}

	fclose(f);
	
	return model;
}



G3DMODEL *g3dmodel_new(void)
{
	G3DMODEL *model;
	
	model= (G3DMODEL *) malloc(sizeof(G3DMODEL));
	assert(model!=NULL);
	
	model->meshCount= 0;
	model->meshes= NULL;

	return model;
}


void g3dmodel_free(G3DMODEL *model)
{
	G3DMESH *mesh;
	int i;
	
	if(model!=NULL)
	{
		if(model->meshes!=NULL)
		{
			for(i= 0; i < model->meshCount; i++)
			{
				mesh= &model->meshes[i];

				if(mesh->vertices!=NULL)
					free(mesh->vertices);
				if(mesh->normals!=NULL)
					free(mesh->normals);
				if(mesh->texCoords!=NULL)
					free(mesh->texCoords);
				if(mesh->colors!=NULL)
					free(mesh->colors);
				if(mesh->indices!=NULL)
					free(mesh->indices);
				
				if(mesh->vertexFrameCount > 1 && mesh->anim_vertices!=NULL)
					free(mesh->anim_vertices);
				if(mesh->normalFrameCount > 1 && mesh->anim_normals!=NULL)
					free(mesh->anim_normals);
				if(mesh->texCoordFrameCount > 1 && mesh->anim_texCoords!=NULL)
					free(mesh->anim_texCoords);
				if(mesh->colorFrameCount > 1 && mesh->anim_colors!=NULL)
					free(mesh->anim_colors);
			}
			
			free(model->meshes);
		}

		free(model);
	}
}


void g3dmodel_display(G3DMODEL *model)
{
	G3DMESH *mesh;
	int i;

	glEnableClientState(GL_VERTEX_ARRAY);
	
	for(i= 0; i < model->meshCount; i++)
	{
		mesh= &model->meshes[i];

	#if 1
		if(mesh->texCoordFrameCount > 0 && mesh->tex!=0)
		{	
			glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

			glBindTexture(GL_TEXTURE_2D, mesh->tex);
			glEnable(GL_TEXTURE_2D);

			glEnableClientState(GL_TEXTURE_COORD_ARRAY);
			glTexCoordPointer(2, GL_FLOAT, 0, mesh->texCoords);
		}
		else
	#endif
		{
			glDisable(GL_TEXTURE_2D);
			glDisableClientState(GL_TEXTURE_COORD_ARRAY);
		}
		
#if 0
		if(mesh->normalFrameCount > 0)
		{
			glEnableClientState(GL_NORMAL_ARRAY);
			glNormalPointer(GL_FLOAT, 0, mesh->normals);
		}
		else
			glDisableClientState(GL_NORMAL_ARRAY);
#endif
		
		glVertexPointer(3, GL_FLOAT, 0, mesh->vertices);
		// glDrawRangeElements(GL_TRIANGLES, 0, mesh->pointCount-1, mesh->indexCount, GL_UNSIGNED_INT, mesh->indices);
		glDrawElements(GL_TRIANGLES, mesh->indexCount, GL_UNSIGNED_INT, mesh->indices);
	}
}


void g3dmodel_anim_display(G3DMODEL *model)
{
	G3DMESH *mesh;
	int i;

	glEnableClientState(GL_VERTEX_ARRAY);
	
	for(i= 0; i < model->meshCount; i++)
	{
		mesh= &model->meshes[i];
	
	#if 1
		if(mesh->texCoordFrameCount > 0 && mesh->tex!=0)
		{	
			glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

			glBindTexture(GL_TEXTURE_2D, mesh->tex);
			glEnable(GL_TEXTURE_2D);

			glEnableClientState(GL_TEXTURE_COORD_ARRAY);
			glTexCoordPointer(2, GL_FLOAT, 0, mesh->anim_texCoords);
		}
		else
	#endif
		{
			glDisable(GL_TEXTURE_2D);
			glDisableClientState(GL_TEXTURE_COORD_ARRAY);
		}
		
#if 0
		if(mesh->normalFrameCount > 0)
		{
			glEnableClientState(GL_NORMAL_ARRAY);
			glNormalPointer(GL_FLOAT, 0, mesh->anim_normals);
		}
		else
			glDisableClientState(GL_NORMAL_ARRAY);
#endif
		
		glVertexPointer(3, GL_FLOAT, 0, mesh->anim_vertices);
		// glDrawRangeElements(GL_TRIANGLES, 0, mesh->pointCount-1, mesh->indexCount, GL_UNSIGNED_INT, mesh->indices);
		glDrawElements(GL_TRIANGLES, mesh->indexCount, GL_UNSIGNED_INT, mesh->indices);
	}
}

void g3dmesh_vertex_anim(G3DMESH *mesh, int timer, int timer_rate, int anim_rate)
{
	VEC3 v;
	VEC3 *a, *b;
	float t, tframe;
	int nframe, frame, i;
	
	t= (float) timer / (float) (anim_rate * timer_rate);
	t= fmod(t, 1.f);
	
	frame= (int) (t * (float) mesh->vertexFrameCount) % mesh->vertexFrameCount;
	nframe= frame +1;
	if(nframe >= mesh->vertexFrameCount)
		// boucler l'animation
		nframe= 0;

	tframe= (float) frame / (float) mesh->vertexFrameCount;
	t= (t - tframe) * (float) mesh->vertexFrameCount;
	
	for(i= 0; i < mesh->pointCount; i++)
	{
		a= &mesh->vertices[frame*mesh->pointCount +i];
		b= &mesh->vertices[nframe*mesh->pointCount +i];
		vec3_sub(v, *b, *a);
		
		vec3_add_mul_const(mesh->anim_vertices[i], *a, t, v);
	}
}

void g3dmesh_normal_anim(G3DMESH *mesh, int timer, int timer_rate, int anim_rate)
{
	//fix
	return;
}

void g3dmesh_tex_anim(G3DMESH *mesh, int timer, int timer_rate, int anim_rate)
{
	VEC2 v;
	VEC2 *a, *b;
	float t, tframe;
	int nframe, frame, i;
	
	t= (float) timer / (float) (anim_rate * timer_rate);
	t= fmod(t, 1.f);
	
	frame= (int) (t * (float) mesh->vertexFrameCount) % mesh->vertexFrameCount;
	nframe= frame +1;
	if(nframe >= mesh->vertexFrameCount)
		nframe= 0;
	
	tframe= (float) frame / (float) mesh->vertexFrameCount;
	t= (t - tframe) * (float) mesh->vertexFrameCount;
	
	for(i= 0; i < mesh->pointCount; i++)
	{
		a= &mesh->texCoords[frame*mesh->pointCount +i];
		b= &mesh->texCoords[nframe*mesh->pointCount +i];
		vec2_sub(v, *b, *a);
		
		vec2_add_mul_const(mesh->anim_texCoords[i], *a, t, v);
	}
}

void g3dmesh_color_anim(G3DMESH *mesh, int timer, int timer_rate, int anim_rate)
{
	//fix
	return;
}


void g3dmesh_anim(G3DMESH *mesh, int timer, int timer_rate, int anim_rate)
{
	if(mesh->vertexFrameCount > 1)
		g3dmesh_vertex_anim(mesh, timer, timer_rate, anim_rate);
	
	if(mesh->normalFrameCount > 1)
		g3dmesh_normal_anim(mesh, timer, timer_rate, anim_rate);

	if(mesh->texCoordFrameCount > 1)
		g3dmesh_tex_anim(mesh, timer, timer_rate, anim_rate);

	if(mesh->colorFrameCount > 1)	
		g3dmesh_color_anim(mesh, timer, timer_rate, anim_rate);
}

void g3dmodel_anim(G3DMODEL *model, int timer, int timer_rate, int anim_rate)
{
	int i;

	for(i= 0; i < model->meshCount; i++)
		g3dmesh_anim(&model->meshes[i], timer, timer_rate, anim_rate);
	
}

