#ifndef _G3DMODEL_H
#define _G3DMODEL_H

#include <stdint.h>
#include <GL/glut.h>

#define STRING_MAX 128
#define PATH_MAX 1024

typedef float VEC[4];

typedef VEC VEC4;
typedef float VEC3[3];
typedef float VEC2[2];

//
typedef struct 
{
	uint32_t vertexFrameCount;
	uint32_t normalFrameCount;
	uint32_t texCoordFrameCount;
	uint32_t colorFrameCount;
	uint32_t pointCount;
	uint32_t indexCount;
	uint8_t hasTexture;
	uint8_t primitive;
	uint8_t cullFace;
	uint8_t texName[64];
} G3DMESH_HEADERV2;

typedef struct
{
	uint32_t vertexFrameCount;
	uint32_t normalFrameCount;
	uint32_t texCoordFrameCount;
	uint32_t colorFrameCount;
	uint32_t pointCount;
	uint32_t indexCount;
	uint32_t properties;
	uint8_t texName[64];
} G3DMESH_HEADER;


typedef struct
{
// mesh data
	GLuint tex;
	char texName[STRING_MAX];

// vertex data counts
	int vertexFrameCount;
	int normalFrameCount;
	int texCoordFrameCount;
	int colorFrameCount;		//per frame, not per vertex
	int pointCount;
	int indexCount;

// vertex data
	VEC3 *vertices;
	VEC3 *normals;
	VEC2 *texCoords;
	VEC4 *colors;
	int *indices;

// anim vertex data
	VEC3 *anim_vertices;
	VEC3 *anim_normals;
	VEC2 *anim_texCoords;
	VEC4 *anim_colors;
	
// properties
	int twoSided;
	int customTexture;
} G3DMESH;


typedef struct
{
	uint8_t id[3];
	uint8_t version;
	uint32_t meshCount;
} G3DMODEL_HEADER;

typedef struct
{
	G3DMESH *meshes;
	int meshCount;
} G3DMODEL;


//
extern void g3dmodel_free(G3DMODEL *model);
extern G3DMODEL *g3dmodel_new(void);
extern G3DMODEL *g3dmodel_load(char *path);

extern void g3dmodel_display(G3DMODEL *model);

extern void g3dmodel_anim_display(G3DMODEL *model);
extern void g3dmodel_anim(G3DMODEL *model, int timer, int timer_rate, int anim_rate);

#endif
