/*
	visualisation objets glest g3d
	http://www.glest.org
	
	mailto:jciehl@bat710.univ-lyon1.fr
 */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <math.h>
#include <unistd.h>

#include <GL/glut.h>

#include "g3dmodel.h"


// globales pour glut
G3DMODEL *scene;

// position et orientation du point de vue
float camera_x, camera_y, camera_z;
float camera_angle, camera_angle2;
float camera_vitesse;

// timer pour l'animation
int timer;
int timer_rate;
int rate_limit;

int anim;
int anim_rate;

// conversion degres vers radians
float radians(float d)
{
	return d * 0.0174532925f;
}

// callback de glut : affichage
void display(void)
{
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glMatrixMode(GL_MODELVIEW);		// selectionne la matrice de la scene
	glLoadIdentity();				// reinitialise les transformations

	// oriente la scene par rapport a la camera qui est restee en 0,0,0
	glRotatef(360. - camera_angle2, 1., 0., 0);
	glRotatef(360. - camera_angle, 0., 1., 0);
	glTranslatef(-camera_x, -camera_y, -camera_z);

#if 0
	// desssine le modele 
	g3dmodel_display(scene);
#else
	// anime et affiche le modele
	g3dmodel_anim(scene, timer, timer_rate, anim_rate);
	g3dmodel_anim_display(scene);
#endif
	
	// termine l'affichage
    glFlush();
	// echange le buffer de presentation et le buffer d'affichage
    glutSwapBuffers();
}

// callback de glut : appellee regulierement
void anime(void)
{
#if 1
	// attente : limite le nombre d'images affichees par seconde
#ifdef _WIN32
	Sleep(1000 / rate_limit);
#else
	usleep(1000000 / rate_limit);
#endif
#endif
	
	// recupere le temps ecoule pour determiner l'image cle de l'animation
	timer= glutGet(GLUT_ELAPSED_TIME);

	// demande un nouvel affichage
	glutPostRedisplay();
}

// callback de glut : gestion du clavier (touches speciales)
void keyboard(int key, int x, int y)
{
	int mod;
	
	mod= glutGetModifiers();
	
	switch(key)
	{
		case 27:
		case 'q':
		case 'Q':
			exit(0);
		break;

		case GLUT_KEY_UP:
			if(mod&GLUT_ACTIVE_CTRL || mod&GLUT_ACTIVE_ALT)
			{
				// rotation de la camera sur X
				camera_angle2+= 1.;
				if(camera_angle2 >= 360.)
					camera_angle2= 0.;
			}
			else
			{
				// deplacement de la camera dans la direction de la visee
				camera_z-= cos(radians(camera_angle)) * camera_vitesse;
				camera_x-= sin(radians(camera_angle)) * camera_vitesse;
			}
		break;

		case GLUT_KEY_DOWN:
			if(mod&GLUT_ACTIVE_CTRL || mod&GLUT_ACTIVE_ALT)
			{
				// rotation de la camera sur Y+
				camera_angle2-= 1.;
				if(camera_angle2 < 0.)
					camera_angle2= 359.;
			}
			else
			{
				// deplacement de la camera dans la direction opposee de la visee
				camera_z+= cos(radians(camera_angle)) * camera_vitesse;
				camera_x+= sin(radians(camera_angle)) * camera_vitesse;
			}
			break;

		case GLUT_KEY_LEFT:
			if(mod&GLUT_ACTIVE_CTRL || mod&GLUT_ACTIVE_ALT)
			{
				// rotation de la camera sur Y-
				camera_angle+= 1.;
				if(camera_angle >= 360.)
					camera_angle= 0.;
			}
			else
			{
				// deplacement perpendiculaire a la direction de visee
				camera_z-= cos(radians(camera_angle + 90.)) * camera_vitesse;
				camera_x-= sin(radians(camera_angle + 90.)) * camera_vitesse;
			}
		break;

		case GLUT_KEY_RIGHT:
			if(mod&GLUT_ACTIVE_CTRL || mod&GLUT_ACTIVE_ALT)
			{
				// rotation de la camera sur Y+
				camera_angle-= 1.;
				if(camera_angle < 0.)
					camera_angle= 359.;
			}
			else
			{
				// deplacement perpendiculaire a la direction de visee
				camera_z-= cos(radians(camera_angle - 90.)) * camera_vitesse;
				camera_x-= sin(radians(camera_angle - 90.)) * camera_vitesse;
			}
		break;

		case GLUT_KEY_PAGE_UP:
			camera_y+= camera_vitesse;
		break;
		
		case GLUT_KEY_PAGE_DOWN:
			camera_y-= camera_vitesse;
		break;
			
		default:
		break;
	}

	// provoque un affichage (appel du callback display)
	glutPostRedisplay();
}

// touches classiques (code ascii) 
void keyboard_char(unsigned char key, int x, int y)
{
	keyboard((int) key, x, y);
}


// mise en place des parametres d'openGL
void initgl(void)
{
    glClearColor(0.0, 0.0, 0.0, 0.0);
	glClearDepth(1.);

	glPolygonMode(GL_FRONT, GL_FILL);
	glPolygonMode(GL_BACK, GL_LINE);

	glEnable(GL_DEPTH_TEST);
	glEnable(GL_CULL_FACE);
	glCullFace(GL_BACK);

	glShadeModel(GL_FLAT);
	
    glMatrixMode(GL_PROJECTION);	// selectionne la matrice de visualisation
    glLoadIdentity();				// reinitialise les transformations

	gluPerspective(50., 1., 1., 10000.);

	camera_x= 0.f;
	camera_y= 0.f;
	camera_z= 5.f;
	camera_angle= 0.f;
	camera_vitesse= 1.f;

	timer= 0;
	timer_rate= 1000;	// le temps ecoule est en millisecondes, cf. fonction anime
	rate_limit= 30;		// frequence d'affichage maximale (pour eviter de consommer tout le cpu)
	
	anim= 0;
	anim_rate= 2;	// duree de la boucle d'animation en secondes
}


int main(int argc, char **argv)
{
	glutInit(&argc, argv);
	
	if(argc!=2)
	{
		printf("usage: %s model.g3d\n", argv[0]);
		return 0;
	}
	
	glutInitDisplayMode (GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
	glutInitWindowSize (600, 600);
	glutInitWindowPosition (100, 100);
	glutCreateWindow ("g3d");
	initgl();

	scene= g3dmodel_load(argv[1]);
	if(scene!=NULL)
	{
		printf("model '%s' done.\n", argv[1]);

		glutKeyboardFunc(keyboard_char);
		glutSpecialFunc(keyboard);
		glutDisplayFunc(display);
		glutIdleFunc(anime);

		glutMainLoop();

		// g3dmodel_free(model);
	}
	else
		printf("model '%s' failed.\n", argv[1]);	
	
	return 0;
}
